import logging

from abc import abstractmethod

from .enums import PNStatusCategory, PNOperationType
from .models.consumer.common import PNStatus
from .models.consumer.objects_v2.channel import PNChannelMetadataResult
from .models.consumer.objects_v2.memberships import PNMembershipResult
from .models.consumer.objects_v2.uuid import PNUUIDMetadataResult
from .models.consumer.pn_error_data import PNErrorData
from .utils import strip_right
from .models.consumer.pubsub import (
    PNPresenceEventResult, PNMessageResult, PNSignalMessageResult, PNMessageActionResult, PNFileMessageResult
)
from .models.server.subscribe import SubscribeMessage, PresenceEnvelope
from .endpoints.file_operations.get_file_url import GetFileDownloadUrl


logger = logging.getLogger("pubnub")


class SubscribeMessageWorker(object):
    TYPE_MESSAGE = 0
    TYPE_SIGNAL = 1
    TYPE_OBJECT = 2
    TYPE_MESSAGE_ACTION = 3
    TYPE_FILE_MESSAGE = 4

    def __init__(self, pubnub_instance, listener_manager_instance, queue_instance, event):
        # assert isinstance(pubnub_instnace, PubNubCore)
        # assert isinstance(listener_manager_instance, ListenerManager)
        # assert isinstance(queue_instance, utils.Queue)

        self._pubnub = pubnub_instance
        self._listener_manager = listener_manager_instance
        self._queue = queue_instance
        self._is_running = None
        self._event = event

    def run(self):
        self._take_message()

    @abstractmethod
    def _take_message(self):
        pass

    def _get_url_for_file_event_message(self, channel, extracted_message):
        return GetFileDownloadUrl(self._pubnub)\
            .channel(channel) \
            .file_name(extracted_message["file"]["name"])\
            .file_id(extracted_message["file"]["id"]).get_complete_url()

    def _process_message(self, message_input):
        if self._pubnub.config.cipher_key is None:
            return message_input
        else:
            try:
                return self._pubnub.config.crypto.decrypt(
                    self._pubnub.config.cipher_key,
                    message_input
                )
            except Exception as exception:
                logger.warning("could not decrypt message: \"%s\", due to error %s" % (message_input, str(exception)))
                pn_status = PNStatus()
                pn_status.category = PNStatusCategory.PNDecryptionErrorCategory
                pn_status.error_data = PNErrorData(str(exception), exception)
                pn_status.error = True
                pn_status.operation = PNOperationType.PNSubscribeOperation
                self._listener_manager.announce_status(pn_status)
                return message_input

    def _process_incoming_payload(self, message):
        assert isinstance(message, SubscribeMessage)

        channel = message.channel
        subscription_match = message.subscription_match
        publish_meta_data = message.publish_metadata

        if channel is not None and channel == subscription_match:
            subscription_match = None

        if "-pnpres" in message.channel:
            presence_payload = PresenceEnvelope.from_json_payload(message.payload)

            stripped_presence_channel = None
            stripped_presence_subscription = None

            if channel is not None:
                stripped_presence_channel = strip_right(channel, "-pnpres")

            if subscription_match is not None:
                stripped_presence_subscription = strip_right(subscription_match, "-pnpres")

            pn_presence_event_result = PNPresenceEventResult(
                event=presence_payload.action,
                channel=stripped_presence_channel,
                subscription=stripped_presence_subscription,
                timetoken=publish_meta_data.publish_timetoken,
                occupancy=presence_payload.occupancy,
                uuid=presence_payload.uuid,
                timestamp=presence_payload.timestamp,
                state=presence_payload.data,
                join=message.payload.get('join', None),
                leave=message.payload.get('leave', None),
                timeout=message.payload.get('timeout', None)
            )
            self._listener_manager.announce_presence(pn_presence_event_result)
        elif message.type == SubscribeMessageWorker.TYPE_OBJECT:
            if message.payload['type'] == 'channel':
                channel_result = PNChannelMetadataResult(
                    event=message.payload['event'],
                    data=message.payload['data']
                )
                self._listener_manager.announce_channel(channel_result)
            elif message.payload['type'] == 'uuid':
                uuid_result = PNUUIDMetadataResult(
                    event=message.payload['event'],
                    data=message.payload['data']
                )
                self._listener_manager.announce_uuid(uuid_result)
            elif message.payload['type'] == 'membership':
                membership_result = PNMembershipResult(
                    event=message.payload['event'],
                    data=message.payload['data']
                )
                self._listener_manager.announce_membership(membership_result)
        elif message.type == SubscribeMessageWorker.TYPE_FILE_MESSAGE:
            extracted_message = self._process_message(message.payload)
            download_url = self._get_url_for_file_event_message(channel, extracted_message)

            pn_file_result = PNFileMessageResult(
                message=extracted_message.get("message"),
                channel=channel,
                subscription=subscription_match,
                timetoken=publish_meta_data.publish_timetoken,
                publisher=message.issuing_client_id,
                file_url=download_url,
                file_id=extracted_message["file"]["id"],
                file_name=extracted_message["file"]["name"]
            )

            self._listener_manager.announce_file_message(pn_file_result)

        else:
            extracted_message = self._process_message(message.payload)
            publisher = message.issuing_client_id

            if extracted_message is None:
                logger.debug("unable to parse payload on #processIncomingMessages")

            if message.type == SubscribeMessageWorker.TYPE_SIGNAL:
                pn_signal_result = PNSignalMessageResult(
                    message=extracted_message,
                    channel=channel,
                    subscription=subscription_match,
                    timetoken=publish_meta_data.publish_timetoken,
                    publisher=publisher
                )
                self._listener_manager.announce_signal(pn_signal_result)

            elif message.type == SubscribeMessageWorker.TYPE_MESSAGE_ACTION:
                message_action = extracted_message['data']
                if 'uuid' not in message_action:
                    message_action['uuid'] = publisher

                message_action_result = PNMessageActionResult(message_action)
                self._listener_manager.announce_message_action(message_action_result)

            else:
                pn_message_result = PNMessageResult(
                    message=extracted_message,
                    channel=channel,
                    subscription=subscription_match,
                    timetoken=publish_meta_data.publish_timetoken,
                    publisher=publisher
                )
                self._listener_manager.announce_message(pn_message_result)



from pubnub.endpoints.entities.endpoint import EntitiesEndpoint, SpaceEndpoint, IncludeCustomEndpoint, \
    CustomAwareEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.entities.space import PNUpdateSpaceResult
from pubnub.utils import write_value_as_string


class UpdateSpace(EntitiesEndpoint, SpaceEndpoint, IncludeCustomEndpoint, CustomAwareEndpoint):
    UPDATE_SPACE_PATH = "/v2/objects/%s/channels/%s"

    def __init__(self, pubnub):
        EntitiesEndpoint.__init__(self, pubnub)
        SpaceEndpoint.__init__(self)
        CustomAwareEndpoint.__init__(self)
        IncludeCustomEndpoint.__init__(self)

        self._name = None
        self._description = None
        self._status = None
        self._type = None

    def space_status(self, space_status):
        self._status = space_status
        self._include_status = True
        return self

    def space_type(self, space_type):
        self._type = space_type
        self._include_type = True
        return self

    def set_name(self, name):
        self._name = str(name)
        return self

    def description(self, description):
        self._description = str(description)
        return self

    def validate_specific_params(self):
        self._validate_space_id()

    def build_path(self):
        return UpdateSpace.UPDATE_SPACE_PATH % (self.pubnub.config.subscribe_key, self._space_id)

    def build_data(self):
        payload = {
            "name": self._name,
            "description": self._description,
            "custom": self._custom
        }
        if self._status:
            payload['status'] = self._status
        if self._type:
            payload['type'] = self._type
        return write_value_as_string(payload)

    def create_response(self, envelope):
        return PNUpdateSpaceResult(envelope)

    def operation_type(self):
        return PNOperationType.PNUpdateSpaceOperation

    def name(self):
        return "Updatea space"

    def http_method(self):
        return HttpMethod.PATCH



from pubnub.endpoints.entities.endpoint import EntitiesEndpoint, SpaceEndpoint, IncludeCustomEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.entities.space import PNFetchSpaceResult


class FetchSpace(EntitiesEndpoint, SpaceEndpoint, IncludeCustomEndpoint):
    FETCH_SPACE_PATH = "/v2/objects/%s/channels/%s"

    def __init__(self, pubnub):
        EntitiesEndpoint.__init__(self, pubnub)
        SpaceEndpoint.__init__(self)
        IncludeCustomEndpoint.__init__(self)

    def build_path(self):
        return FetchSpace.FETCH_SPACE_PATH % (self.pubnub.config.subscribe_key, self._space_id)

    def validate_specific_params(self):
        self._validate_space_id()

    def create_response(self, envelope):
        return PNFetchSpaceResult(envelope)

    def operation_type(self):
        return PNOperationType.PNFetchSpaceOperation

    def name(self):
        return "Fetch Space"

    def http_method(self):
        return HttpMethod.GET



from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.models.consumer.message_actions import PNGetMessageActionsResult, PNMessageAction
from pubnub.enums import HttpMethod, PNOperationType


class GetMessageActions(Endpoint):
    GET_MESSAGE_ACTIONS_PATH = '/v1/message-actions/%s/channel/%s'
    MAX_LIMIT = 100

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._channel = None
        self._start = None
        self._end = None
        self._limit = GetMessageActions.MAX_LIMIT

    def channel(self, channel):
        self._channel = str(channel)
        return self

    def start(self, start):
        assert isinstance(start, str)
        self._start = start
        return self

    def end(self, end):
        assert isinstance(end, str)
        self._end = end
        return self

    def limit(self, limit):
        assert isinstance(limit, str)
        self._limit = limit
        return self

    def custom_params(self):
        params = {}

        if self._start is not None:
            params['start'] = self._start

        if self._end is not None and self._start is None:
            params['end'] = self._end

        if self._limit != GetMessageActions.MAX_LIMIT:
            params['limit'] = self._limit

        return params

    def build_path(self):
        return GetMessageActions.GET_MESSAGE_ACTIONS_PATH % (
            self.pubnub.config.subscribe_key,
            utils.url_encode(self._channel)
        )

    def http_method(self):
        return HttpMethod.GET

    def is_auth_required(self):
        return True

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_channel()

    def create_response(self, envelope):   # pylint: disable=W0221
        result = envelope
        result['actions'] = []
        for action in result['data']:
            result['actions'].append(PNMessageAction(action))

        return PNGetMessageActionsResult(result)

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNGetMessageActions

    def name(self):
        return 'Get message actions'



import hashlib
import json
import random
from base64 import decodebytes, encodebytes

from .crypto_core import PubNubCrypto
from Cryptodome.Cipher import AES
from Cryptodome.Util.Padding import pad, unpad


Initial16bytes = '0123456789012345'


class PubNubCryptodome(PubNubCrypto):
    def __init__(self, pubnub_config):
        self.pubnub_configuration = pubnub_config

    def encrypt(self, key, msg, use_random_iv=False):
        secret = self.get_secret(key)
        initialization_vector = self.get_initialization_vector(use_random_iv)

        cipher = AES.new(bytes(secret[0:32], 'utf-8'), AES.MODE_CBC, bytes(initialization_vector, 'utf-8'))
        encrypted_message = cipher.encrypt(self.pad(msg.encode('utf-8')))
        msg_with_iv = self.append_random_iv(encrypted_message, use_random_iv, bytes(initialization_vector, "utf-8"))

        return encodebytes(msg_with_iv).decode('utf-8').replace("\n", "")

    def decrypt(self, key, msg, use_random_iv=False):
        secret = self.get_secret(key)

        decoded_message = decodebytes(msg.encode("utf-8"))
        initialization_vector, extracted_message = self.extract_random_iv(decoded_message, use_random_iv)
        cipher = AES.new(bytes(secret[0:32], "utf-8"), AES.MODE_CBC, initialization_vector)
        plain = self.depad((cipher.decrypt(extracted_message)).decode('utf-8'))

        try:
            return json.loads(plain)
        except Exception:
            return plain

    def append_random_iv(self, message, use_random_iv, initialization_vector):
        if self.pubnub_configuration.use_random_initialization_vector or use_random_iv:
            return initialization_vector + message
        else:
            return message

    def extract_random_iv(self, message, use_random_iv):
        if self.pubnub_configuration.use_random_initialization_vector or use_random_iv:
            return message[0:16], message[16:]
        else:
            return bytes(Initial16bytes, "utf-8"), message

    def get_initialization_vector(self, use_random_iv):
        if self.pubnub_configuration.use_random_initialization_vector or use_random_iv:
            return "{0:016}".format(random.randint(0, 9999999999999999))
        else:
            return Initial16bytes

    def pad(self, msg, block_size=16):
        padding = block_size - (len(msg) % block_size)
        return msg + (chr(padding) * padding).encode('utf-8')

    def depad(self, msg):
        return msg[0:-ord(msg[-1])]

    def get_secret(self, key):
        return hashlib.sha256(key.encode("utf-8")).hexdigest()


class PubNubFileCrypto(PubNubCryptodome):
    def encrypt(self, key, file):
        secret = self.get_secret(key)
        initialization_vector = self.get_initialization_vector(use_random_iv=True)
        cipher = AES.new(bytes(secret[0:32], "utf-8"), AES.MODE_CBC, bytes(initialization_vector, 'utf-8'))
        initialization_vector = bytes(initialization_vector, 'utf-8')

        return self.append_random_iv(
            cipher.encrypt(pad(file, 16)),
            use_random_iv=True,
            initialization_vector=initialization_vector
        )

    def decrypt(self, key, file):
        secret = self.get_secret(key)
        initialization_vector, extracted_file = self.extract_random_iv(file, use_random_iv=True)
        cipher = AES.new(bytes(secret[0:32], "utf-8"), AES.MODE_CBC, initialization_vector)

        return unpad(cipher.decrypt(extracted_file), 16)



from pubnub.endpoints.file_operations.file_based_endpoint import FileOperationEndpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub import utils
from pubnub.models.consumer.file import PNGetFileDownloadURLResult


class GetFileDownloadUrl(FileOperationEndpoint):
    GET_FILE_DOWNLOAD_URL = "/v1/files/%s/channels/%s/files/%s/%s"

    def __init__(self, pubnub, file_name=None, file_id=None):
        FileOperationEndpoint.__init__(self, pubnub)
        self._file_id = file_id
        self._file_name = file_name

    def build_path(self):
        return GetFileDownloadUrl.GET_FILE_DOWNLOAD_URL % (
            self.pubnub.config.subscribe_key,
            utils.url_encode(self._channel),
            self._file_id,
            self._file_name
        )

    def get_complete_url(self):
        endpoint_options = self.options()
        endpoint_options.merge_params_in(self.custom_params())
        query_params = '?' + endpoint_options.query_string

        return self.pubnub.config.scheme_extended() + self.pubnub.base_origin + self.build_path() + query_params

    def file_id(self, file_id):
        self._file_id = file_id
        return self

    def file_name(self, file_name):
        self._file_name = file_name
        return self

    def http_method(self):
        return HttpMethod.GET

    def custom_params(self):
        return {}

    def is_auth_required(self):
        return True

    def non_json_response(self):
        return True

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_channel()
        self.validate_file_id()
        self.validate_file_name()

    def create_response(self, envelope, data=None):
        return PNGetFileDownloadURLResult(envelope)

    def operation_type(self):
        return PNOperationType.PNGetFileDownloadURLAction

    def allow_redirects(self):
        return False

    def name(self):
        return "Get file download url"



from pubnub import utils
from pubnub.endpoints.objects_v2.objects_endpoint import ObjectsEndpoint, ListEndpoint, \
    IncludeCustomEndpoint, UuidEndpoint, ChannelIncludeEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.objects_v2.memberships import PNRemoveMembershipsResult


class RemoveMemberships(ObjectsEndpoint, UuidEndpoint, ListEndpoint, IncludeCustomEndpoint,
                        ChannelIncludeEndpoint):
    REMOVE_MEMBERSHIPS_PATH = "/v2/objects/%s/uuids/%s/channels"

    def __init__(self, pubnub):
        ObjectsEndpoint.__init__(self, pubnub)
        ListEndpoint.__init__(self)
        UuidEndpoint.__init__(self)
        IncludeCustomEndpoint.__init__(self)
        ChannelIncludeEndpoint.__init__(self)

        self._channel_memberships = []

    def channel_memberships(self, channel_memberships):
        self._channel_memberships = list(channel_memberships)
        return self

    def build_path(self):
        return RemoveMemberships.REMOVE_MEMBERSHIPS_PATH % (self.pubnub.config.subscribe_key, self._effective_uuid())

    def build_data(self):
        channel_memberships_to_delete = []

        for channel_membership in self._channel_memberships:
            channel_memberships_to_delete.append(channel_membership.to_payload_dict())

        payload = {
            "set": [],
            "delete": channel_memberships_to_delete
        }
        return utils.write_value_as_string(payload)

    def validate_specific_params(self):
        self._validate_uuid()

    def create_response(self, envelope):
        return PNRemoveMembershipsResult(envelope)

    def operation_type(self):
        return PNOperationType.PNRemoveMembershipsOperation

    def name(self):
        return "Remove Memberships"

    def http_method(self):
        return HttpMethod.PATCH



from pubnub.endpoints.file_operations.file_based_endpoint import FileOperationEndpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.crypto import PubNubFileCrypto
from pubnub.models.consumer.file import PNDownloadFileResult
from pubnub.request_handlers.requests_handler import RequestsRequestHandler
from pubnub.endpoints.file_operations.get_file_url import GetFileDownloadUrl


class DownloadFileNative(FileOperationEndpoint):
    def __init__(self, pubnub):
        FileOperationEndpoint.__init__(self, pubnub)
        self._file_id = None
        self._file_name = None
        self._pubnub = pubnub
        self._download_data = None
        self._cipher_key = None

    def cipher_key(self, cipher_key):
        self._cipher_key = cipher_key
        return self

    def build_path(self):
        return self._download_data.result.file_url

    def http_method(self):
        return HttpMethod.GET

    def is_auth_required(self):
        return False

    def custom_params(self):
        return {}

    def file_id(self, file_id):
        self._file_id = file_id
        return self

    def file_name(self, file_name):
        self._file_name = file_name
        return self

    def decrypt_payload(self, data):
        return PubNubFileCrypto(self._pubnub.config).decrypt(
            self._cipher_key or self._pubnub.config.cipher_key,
            data
        )

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_channel()
        self.validate_file_name()
        self.validate_file_id()

    def create_response(self, envelope):
        if self._cipher_key or self._pubnub.config.cipher_key:
            return PNDownloadFileResult(self.decrypt_payload(envelope.content))
        else:
            return PNDownloadFileResult(envelope.content)

    def non_json_response(self):
        return True

    def operation_type(self):
        return PNOperationType.PNDownloadFileAction

    def use_base_path(self):
        return False

    def build_params_callback(self):
        return lambda a: {}

    def name(self):
        return "Downloading file"

    def sync(self):
        self._download_data = GetFileDownloadUrl(self._pubnub)\
            .channel(self._channel)\
            .file_name(self._file_name)\
            .file_id(self._file_id)\
            .sync()

        return super(DownloadFileNative, self).sync()

    def pn_async(self, callback):
        return RequestsRequestHandler(self._pubnub).async_file_based_operation(self.sync, callback, "File Download")



from abc import ABCMeta, abstractmethod

import logging
import zlib

from pubnub import utils
from pubnub.enums import PNStatusCategory, HttpMethod
from pubnub.errors import (
    PNERR_SUBSCRIBE_KEY_MISSING, PNERR_PUBLISH_KEY_MISSING, PNERR_CHANNEL_OR_GROUP_MISSING,
    PNERR_SECRET_KEY_MISSING, PNERR_CHANNEL_MISSING, PNERR_FILE_OBJECT_MISSING,
    PNERR_FILE_ID_MISSING, PNERR_FILE_NAME_MISSING
)
from pubnub.exceptions import PubNubException
from pubnub.models.consumer.common import PNStatus
from pubnub.models.consumer.pn_error_data import PNErrorData
from ..structures import RequestOptions, ResponseInfo

logger = logging.getLogger("pubnub")


class Endpoint(object):
    SERVER_RESPONSE_SUCCESS = 200
    SERVER_RESPONSE_FORBIDDEN = 403
    SERVER_RESPONSE_BAD_REQUEST = 400

    __metaclass__ = ABCMeta
    _path = None

    def __init__(self, pubnub):
        self.pubnub = pubnub
        self._cancellation_event = None
        self._sort_params = False
        self._use_compression = self.pubnub.config.should_compress

    def cancellation_event(self, event):
        self._cancellation_event = event
        return self

    @abstractmethod
    def build_path(self):
        pass

    @abstractmethod
    def custom_params(self):
        raise NotImplementedError

    def build_data(self):
        return None

    @abstractmethod
    def http_method(self):
        pass

    @abstractmethod
    def validate_params(self):
        pass

    @abstractmethod
    def create_response(self, endpoint):
        pass

    @abstractmethod
    def operation_type(self):
        raise NotImplementedError

    @abstractmethod
    def name(self):
        pass

    @abstractmethod
    def request_timeout(self):
        pass

    @abstractmethod
    def connect_timeout(self):
        pass

    def is_auth_required(self):
        raise NotImplementedError

    def affected_channels(self):
        return None

    def affected_channels_groups(self):
        return None

    def allow_redirects(self):
        return True

    def use_base_path(self):
        return True

    def is_compressable(self):
        return False

    def request_headers(self):
        headers = {}
        if self.__compress_request():
            headers["Content-Encoding"] = "gzip"
        if self.http_method() == HttpMethod.POST:
            headers["Content-type"] = "application/json"

        return headers

    def build_file_upload_request(self):
        return

    def non_json_response(self):
        return False

    def encoded_params(self):
        return {}

    def get_path(self):
        if not self._path:
            self._path = self.build_path()
        return self._path

    def options(self):
        data = self.build_data()
        if data and self.__compress_request():
            data = zlib.compress(data.encode('utf-8'), level=2)
        return RequestOptions(
            path=self.get_path(),
            params_callback=self.build_params_callback(),
            method=self.http_method(),
            request_timeout=self.request_timeout(),
            connect_timeout=self.connect_timeout(),
            create_response=self.create_response,
            create_status=self.create_status,
            create_exception=self.create_exception,
            operation_type=self.operation_type(),
            data=data,
            files=self.build_file_upload_request(),
            sort_arguments=self._sort_params,
            allow_redirects=self.allow_redirects(),
            use_base_path=self.use_base_path(),
            request_headers=self.request_headers(),
            non_json_response=self.non_json_response()
        )

    def sync(self):
        self.validate_params()
        envelope = self.pubnub.request_sync(self.options())

        if envelope.status.is_error():
            raise envelope.status.error_data.exception

        return envelope

    def pn_async(self, callback):
        try:
            self.validate_params()
            options = self.options()
        except PubNubException as e:
            callback(None, self.create_status(PNStatusCategory.PNBadRequestCategory, None, None, e))
            return

        def callback_wrapper(envelope):
            callback(envelope.result, envelope.status)

        return self.pubnub.request_async(endpoint_name=self.name(),
                                         endpoint_call_options=options,
                                         callback=callback_wrapper,
                                         # REVIEW: include self._cancellation_event into options?
                                         cancellation_event=self._cancellation_event)

    def result(self):
        def handler():
            self.validate_params()
            return self.options()

        return self.pubnub.request_result(options_func=handler,
                                          cancellation_event=self._cancellation_event)

    def future(self):
        def handler():
            self.validate_params()
            return self.options()

        return self.pubnub.request_future(options_func=handler,
                                          cancellation_event=self._cancellation_event)

    def deferred(self):
        def handler():
            self.validate_params()
            return self.options()

        return self.pubnub.request_deferred(options_func=handler,
                                            cancellation_event=self._cancellation_event)

    def build_params_callback(self):
        def callback(params_to_merge):
            custom_params = self.custom_params()
            custom_params.update(params_to_merge)

            custom_params['pnsdk'] = self.pubnub.sdk_name
            custom_params['uuid'] = self.pubnub.uuid

            for query_key, query_value in self.pubnub._telemetry_manager.operation_latencies().items():
                custom_params[query_key] = query_value

            if self.is_auth_required():
                if self.pubnub._get_token():
                    custom_params["auth"] = self.pubnub._get_token()
                elif self.pubnub.config.auth_key:
                    custom_params["auth"] = self.pubnub.config.auth_key

            if self.pubnub.config.secret_key:
                utils.sign_request(self, self.pubnub, custom_params, self.http_method(), self.build_data())

            custom_params.update(self.encoded_params())

            # reassign since pnsdk should be signed unencoded
            custom_params['pnsdk'] = utils.url_encode(self.pubnub.sdk_name)

            return custom_params

        return callback

    def validate_subscribe_key(self):
        if self.pubnub.config.subscribe_key is None or len(self.pubnub.config.subscribe_key) == 0:
            raise PubNubException(pn_error=PNERR_SUBSCRIBE_KEY_MISSING)

    def validate_secret_key(self):
        if self.pubnub.config.secret_key is None or len(self.pubnub.config.secret_key) == 0:
            raise PubNubException(pn_error=PNERR_SECRET_KEY_MISSING)

    def validate_channel(self):
        if self._channel is None or len(self._channel) == 0:
            raise PubNubException(pn_error=PNERR_CHANNEL_MISSING)

    def validate_channels_and_groups(self):
        if len(self._channels) == 0 and len(self._groups) == 0:
            raise PubNubException(pn_error=PNERR_CHANNEL_OR_GROUP_MISSING)

    def validate_publish_key(self):
        if self.pubnub.config.publish_key is None or len(self.pubnub.config.publish_key) == 0:
            raise PubNubException(pn_error=PNERR_PUBLISH_KEY_MISSING)

    def validate_file_object(self):
        if not self._file_object:
            raise PubNubException(pn_error=PNERR_FILE_OBJECT_MISSING)

    def validate_file_name(self):
        if not self._file_name:
            raise PubNubException(pn_error=PNERR_FILE_NAME_MISSING)

    def validate_file_id(self):
        if not self._file_id:
            raise PubNubException(pn_error=PNERR_FILE_ID_MISSING)

    def create_status(self, category, response, response_info, exception):
        if response_info is not None:
            assert isinstance(response_info, ResponseInfo)

        pn_status = PNStatus()

        if response is None or exception is not None:
            pn_status.error = True

        if response is not None:
            pn_status.original_response = response

        if exception is not None:
            pn_status.error_data = PNErrorData(str(exception), exception)

        if response_info is not None:
            pn_status.status_code = response_info.status_code
            pn_status.tls_enabled = response_info.tls_enabled
            pn_status.origin = response_info.origin
            pn_status.uuid = response_info.uuid
            pn_status.auth_key = response_info.auth_key
            pn_status.client_request = response_info.client_request
            pn_status.client_response = response_info.client_response

        pn_status.operation = self.operation_type()
        pn_status.category = category
        pn_status.affected_channels = self.affected_channels()
        pn_status.affected_groups = self.affected_channels_groups()

        return pn_status

    """ Used by asyncio and tornado clients to build exceptions

    The only difference with create_status() method is that a status
    is wrapped with an exception and also contains this exception inside
    as 'status.error_data.exception'
    """

    def create_exception(self, category, response, response_info, exception):
        status = self.create_status(category, response, response_info, exception)

        exception.status = status

        return exception

    def __compress_request(self):
        return (self.is_compressable() and self._use_compression)



class SubscribeOperation(object):
    def __init__(self, channels=None, channel_groups=None, presence_enabled=None, timetoken=None):
        assert isinstance(channels, (list, tuple))
        assert isinstance(channel_groups, (list, tuple))
        assert isinstance(presence_enabled, bool)
        assert isinstance(timetoken, int)

        self.channels = channels
        self.channel_groups = channel_groups
        self.presence_enabled = presence_enabled
        self.timetoken = timetoken


class UnsubscribeOperation(object):
    def __init__(self, channels=None, channel_groups=None):
        assert isinstance(channels, (list, tuple))
        assert isinstance(channel_groups, (list, tuple))

        self.channels = channels
        self.channel_groups = channel_groups


class StateOperation(object):
    def __init__(self, channels=None, channel_groups=None, state=None):
        self.channels = channels
        self.channel_groups = channel_groups
        self.state = state



from pubnub.models.consumer.message_actions import PNMessageAction


class PNMessageResult(object):
    def __init__(self, message, subscription, channel, timetoken, user_metadata=None, publisher=None):

        if subscription is not None:
            assert isinstance(subscription, str)

        if channel is not None:
            assert isinstance(channel, str)

        if publisher is not None:
            assert isinstance(publisher, str)

        assert isinstance(timetoken, int)

        if user_metadata is not None:
            assert isinstance(user_metadata, object)

        self.message = message
        # DEPRECATED: subscribed_channel and actual_channel properties are deprecated
        # self.subscribed_channel = subscribed_channel <= now known as subscription
        # self.actual_channel = actual_channel <= now known as channel

        self.channel = channel
        self.subscription = subscription

        self.timetoken = timetoken
        self.user_metadata = user_metadata
        self.publisher = publisher


class PNSignalMessageResult(PNMessageResult):
    pass


class PNFileMessageResult(PNMessageResult):
    def __init__(
            self, message, subscription,
            channel, timetoken, publisher,
            file_url, file_id, file_name
    ):
        super(PNFileMessageResult, self).__init__(message, subscription, channel, timetoken, publisher=publisher)
        self.file_url = file_url
        self.file_id = file_id
        self.file_name = file_name


class PNPresenceEventResult(object):
    def __init__(self, event, uuid, timestamp, occupancy, subscription, channel,
                 timetoken, state, join, leave, timeout, user_metadata=None):

        assert isinstance(event, str)
        assert isinstance(timestamp, int)
        assert isinstance(occupancy, int)
        assert isinstance(channel, str)
        assert isinstance(timetoken, int)

        if user_metadata is not None:
            assert isinstance(user_metadata, object)

        if state is not None:
            assert isinstance(state, dict)

        self.event = event
        self.uuid = uuid
        self.timestamp = timestamp
        self.occupancy = occupancy
        self.state = state
        self.join = join
        self.leave = leave
        self.timeout = timeout

        # DEPRECATED: subscribed_channel and actual_channel properties are deprecated
        # self.subscribed_channel = subscribed_channel <= now known as subscription
        # self.actual_channel = actual_channel <= now known as channel
        self.subscription = subscription
        self.channel = channel

        self.timetoken = timetoken
        self.user_metadata = user_metadata


class PNMessageActionResult(PNMessageAction):

    def __init__(self, result):
        super(PNMessageActionResult, self).__init__(result)


class PNPublishResult(object):
    def __init__(self, envelope, timetoken):
        """
        Representation of publish server response

        :param timetoken: of publish operation
        """
        self.timetoken = timetoken

    def __str__(self):
        return "Publish success with timetoken %s" % self.timetoken


class PNFireResult(object):
    def __init__(self, envelope, timetoken):
        """
        Representation of fire server response

        :param timetoken: of fire operation
        """
        self.timetoken = timetoken

    def __str__(self):
        return "Fire success with timetoken %s" % self.timetoken



from pubnub import utils
from pubnub.endpoints.objects_v2.objects_endpoint import ObjectsEndpoint, UuidEndpoint, \
    IncludeCustomEndpoint, CustomAwareEndpoint
from pubnub.enums import PNOperationType
from pubnub.enums import HttpMethod
from pubnub.models.consumer.objects_v2.uuid import PNSetUUIDMetadataResult


class SetUuid(ObjectsEndpoint, UuidEndpoint, IncludeCustomEndpoint, CustomAwareEndpoint):
    SET_UID_PATH = "/v2/objects/%s/uuids/%s"

    def __init__(self, pubnub):
        ObjectsEndpoint.__init__(self, pubnub)
        UuidEndpoint.__init__(self)
        IncludeCustomEndpoint.__init__(self)
        CustomAwareEndpoint.__init__(self)

        self._name = None
        self._email = None
        self._external_id = None
        self._profile_url = None

    def set_name(self, name):
        self._name = str(name)
        return self

    def email(self, email):
        self._email = str(email)
        return self

    def external_id(self, external_id):
        self._external_id = str(external_id)
        return self

    def profile_url(self, profile_url):
        self._profile_url = str(profile_url)
        return self

    def build_path(self):
        return SetUuid.SET_UID_PATH % (self.pubnub.config.subscribe_key, self._effective_uuid())

    def build_data(self):
        payload = {
            "name": self._name,
            "email": self._email,
            "externalId": self._external_id,
            "profileUrl": self._profile_url,
            "custom": self._custom
        }
        return utils.write_value_as_string(payload)

    def validate_specific_params(self):
        self._validate_uuid()

    def create_response(self, envelope):
        return PNSetUUIDMetadataResult(envelope)

    def operation_type(self):
        return PNOperationType.PNSetUuidMetadataOperation

    def name(self):
        return "Set UUID"

    def http_method(self):
        return HttpMethod.PATCH



from abc import abstractmethod, ABCMeta

from pubnub.models.consumer.objects_v2.page import PNPageable


class PNChannelMembership:
    __metaclass__ = ABCMeta

    def __init__(self, channel):
        self._channel = channel

    @staticmethod
    def channel(channel):
        return JustChannel(channel)

    @staticmethod
    def channel_with_custom(channel, custom):
        return ChannelWithCustom(channel, custom)

    @abstractmethod
    def to_payload_dict(self):
        return None


class JustChannel(PNChannelMembership):
    def __init__(self, channel):
        PNChannelMembership.__init__(self, channel)

    def to_payload_dict(self):
        return {
            "channel": {
                "id": str(self._channel)
            }
        }


class ChannelWithCustom(PNChannelMembership):
    def __init__(self, channel, custom):
        PNChannelMembership.__init__(self, channel)
        self._custom = custom

    def to_payload_dict(self):
        return {
            "channel": {
                "id": str(self._channel)
            },
            "custom": dict(self._custom)
        }


class PNSetMembershipsResult(PNPageable):
    def __init__(self, result):
        PNPageable.__init__(self, result)
        self.data = result["data"]
        self.status = result["status"]

    def __str__(self):
        return "Set Memberships metatdata: %s" % self.data


class PNGetMembershipsResult(PNPageable):
    def __init__(self, result):
        PNPageable.__init__(self, result)
        self.data = result["data"]
        self.status = result["status"]

    def __str__(self):
        return "Get Memberships metatdata: %s" % self.data


class PNRemoveMembershipsResult(PNPageable):
    def __init__(self, result):
        PNPageable.__init__(self, result)
        self.data = result["data"]
        self.status = result["status"]

    def __str__(self):
        return "Remove Memberships metatdata: %s" % self.data


class PNManageMembershipsResult(PNPageable):
    def __init__(self, result):
        PNPageable.__init__(self, result)
        self.data = result["data"]
        self.status = result["status"]

    def __str__(self):
        return "Manage Channel Members metatdata: %s" % self.data


class PNMembershipResult(object):
    def __init__(self, event, data):
        self.data = data
        self.event = event

    def __str__(self):
        return "Membership %s event with data: %s" % (self.event, self.data)



from pubnub.endpoints.entities.endpoint import EntitiesEndpoint, IncludeCustomEndpoint, ListEndpoint, SpaceEndpoint, \
    UserEndpoint
from pubnub.enums import PNOperationType, HttpMethod
from pubnub.models.consumer.entities.membership import PNSpaceMembershipsResult, PNUserMembershipsResult


class FetchUserMemberships(EntitiesEndpoint, IncludeCustomEndpoint, UserEndpoint, ListEndpoint):
    MEMBERSHIP_PATH = "/v2/objects/%s/uuids/%s/channels"

    def __init__(self, pubnub):
        EntitiesEndpoint.__init__(self, pubnub)
        ListEndpoint.__init__(self)
        IncludeCustomEndpoint.__init__(self)
        UserEndpoint.__init__(self)

    def build_path(self):
        return FetchUserMemberships.MEMBERSHIP_PATH % (self.pubnub.config.subscribe_key, self._user_id)

    def validate_specific_params(self):
        self._validate_user_id()

    def create_response(self, envelope):
        return PNUserMembershipsResult(envelope)

    def operation_type(self):
        return PNOperationType.PNFetchUserMembershipsOperation

    def name(self):
        return "Fetch User Memberships"

    def http_method(self):
        return HttpMethod.GET


class FetchSpaceMemberships(EntitiesEndpoint, IncludeCustomEndpoint, SpaceEndpoint):
    MEMBERSHIP_PATH = "/v2/objects/%s/channels/%s/uuids"

    def __init__(self, pubnub):
        EntitiesEndpoint.__init__(self, pubnub)
        ListEndpoint.__init__(self)
        IncludeCustomEndpoint.__init__(self)
        UserEndpoint.__init__(self)

    def build_path(self):
        return FetchSpaceMemberships.MEMBERSHIP_PATH % (self.pubnub.config.subscribe_key, self._space_id)

    def validate_specific_params(self):
        self._validate_space_id()

    def create_response(self, envelope):
        return PNSpaceMembershipsResult(envelope)

    def operation_type(self):
        return PNOperationType.PNFetchSpaceMembershipsOperation

    def name(self):
        return "Fetch Space Memberships"

    def http_method(self):
        return HttpMethod.GET



class PNMessageCountResult(object):
    def __init__(self, result):
        """
        Representation of message count server response

        :param result: result of message count operation
        """
        self._result = result
        self.channels = result['channels']

    def __str__(self):
        return "Message count for channels: {}".format(self.channels)



class PubNubException(Exception):
    def __init__(self, errormsg="", status_code=0, pn_error=None, status=None):
        self._errormsg = errormsg
        self._status_code = status_code
        self._pn_error = pn_error
        self.status = status

        if len(str(errormsg)) > 0 and int(status_code) > 0:
            msg = str(pn_error) + " (" + str(status_code) + "): " + str(errormsg)
        elif len(str(errormsg)) > 0:
            msg = str(pn_error) + ": " + str(errormsg)
        else:
            msg = str(pn_error)

        super(PubNubException, self).__init__(msg)

    @property
    def _status(self):
        raise DeprecationWarning
        return self.status



"""
Possible responses of PAMv3 request
"""


class _PAMv3Result(object):
    def __init__(self, token):
        self.token = token

    @classmethod
    def from_json(cls, json_input):
        return cls(
            token=json_input['token']
        )


class PNGrantTokenResult(_PAMv3Result):
    def __str__(self):
        return "Grant token: %s" % \
               (self.token)

    def get_token(self):
        return self.token


class PNRevokeTokenResult:
    def __init__(self, result):
        self.status = result['status']

    def __str__(self):
        return "Revoke token success with status: %s" % self.status



from pubnub.models.consumer.objects_v2.page import PNPageable


class PNSetChannelMetadataResult(object):
    def __init__(self, result):
        self.data = result["data"]
        self.status = result["status"]

    def __str__(self):
        return "Set Channel metatdata: %s" % self.data


class PNGetChannelMetadataResult(object):
    def __init__(self, result):
        self.data = result["data"]
        self.status = result["status"]

    def __str__(self):
        return "Get Channel metatdata: %s" % self.data


class PNRemoveChannelMetadataResult(object):
    def __init__(self, result):
        self.data = result["data"]
        self.status = result["status"]

    def __str__(self):
        return "Get Channel metatdata: %s" % self.data


class PNGetAllChannelMetadataResult(PNPageable):
    def __init__(self, result):
        PNPageable.__init__(self, result)
        self.data = result["data"]
        self.status = result["status"]

    def __str__(self):
        return "Get all Channel metatdata: %s" % self.data


class PNChannelMetadataResult(object):
    def __init__(self, event, data):
        self.data = data
        self.event = event

    def __str__(self):
        return "Channel %s event with data: %s" % (self.event, self.data)



from pubnub.models.consumer.entities.result import PNEntityPageableResult


class PNMembershipsResult(PNEntityPageableResult):
    _description = "Set Memberships: %s"

    def __init__(self, result):
        super().__init__(result)
        self.status = result["status"]

    def rename_channel(result):
        result['space'] = result.pop('channel')
        return result

    def rename_uuid(result):
        result['user'] = result.pop('uuid')
        return result


class PNUserMembershipsResult(PNMembershipsResult):
    def __init__(self, result):
        super().__init__(result)
        self.data = [PNMembershipsResult.rename_channel(space) for space in result['data']]


class PNSpaceMembershipsResult(PNMembershipsResult):
    def __init__(self, result):
        super().__init__(result)
        self.data = [PNMembershipsResult.rename_uuid(user) for user in result['data']]



from abc import ABCMeta, abstractmethod


class PNCallback(object):
    @abstractmethod
    def on_response(self, x, status):
        pass


class SubscribeCallback(object):
    __metaclass__ = ABCMeta

    @abstractmethod
    def status(self, pubnub, status):
        pass

    @abstractmethod
    def message(self, pubnub, message):
        pass

    @abstractmethod
    def presence(self, pubnub, presence):
        pass

    def signal(self, pubnub, signal):
        pass

    def channel(self, pubnub, channel):
        pass

    def uuid(self, pubnub, uuid):
        pass

    def membership(self, pubnub, membership):
        pass

    def message_action(self, pubnub, message_action):
        pass

    def file(self, pubnub, file_message):
        pass


class ReconnectionCallback(object):
    @abstractmethod
    def on_reconnect(self):
        pass



class PNChannelGroupsAddChannelResult(object):
    def __str__(self):
        return "Channel successfully added"


class PNChannelGroupsRemoveChannelResult(object):
    def __str__(self):
        return "Channel successfully removed"


class PNChannelGroupsRemoveGroupResult(object):
    def __str__(self):
        return "Group successfully removed"


class PNChannelGroupsListResult(object):
    def __init__(self, channels):
        self.channels = channels

    def __str__(self):
        return "Group contains following channels: %s" % ", ".join(self.channels)



from pubnub import utils
from pubnub.endpoints.endpoint import Endpoint
from pubnub.enums import HttpMethod, PNOperationType
from pubnub.models.consumer.presence import PNWhereNowResult
from pubnub.endpoints.mixins import UUIDValidatorMixin


class WhereNow(Endpoint, UUIDValidatorMixin):
    # /v2/presence/sub-key/<subscribe_key>/uuid/<uuid>
    WHERE_NOW_PATH = "/v2/presence/sub-key/%s/uuid/%s"

    def __init__(self, pubnub):
        Endpoint.__init__(self, pubnub)
        self._uuid = pubnub.config.uuid

    def uuid(self, uuid):
        self._uuid = uuid
        return self

    def custom_params(self):
        return {}

    def build_path(self):
        return WhereNow.WHERE_NOW_PATH % (self.pubnub.config.subscribe_key, utils.url_encode(self._uuid))

    def http_method(self):
        return HttpMethod.GET

    def validate_params(self):
        self.validate_subscribe_key()
        self.validate_uuid()

    def is_auth_required(self):
        return True

    def create_response(self, envelope):
        return PNWhereNowResult.from_json(envelope)

    def request_timeout(self):
        return self.pubnub.config.non_subscribe_request_timeout

    def connect_timeout(self):
        return self.pubnub.config.connect_timeout

    def operation_type(self):
        return PNOperationType.PNWhereNowOperation

    def name(self):
        return "WhereNow"



import aiohttp

from pubnub.endpoints.file_operations.send_file import SendFileNative
from pubnub.endpoints.file_operations.publish_file_message import PublishFileMessage
from pubnub.endpoints.file_operations.fetch_upload_details import FetchFileUploadS3Data


class AsyncioSendFile(SendFileNative):
    def build_file_upload_request(self):
        file = self.encrypt_payload()
        form_data = aiohttp.FormData()
        for form_field in self._file_upload_envelope.result.data["form_fields"]:
            form_data.add_field(form_field["key"], form_field["value"], content_type="multipart/form-data")
        form_data.add_field("file", file, filename=self._file_name, content_type="application/octet-stream")

        return form_data

    def options(self):
        request_options = super(SendFileNative, self).options()
        request_options.data = request_options.files
        return request_options

    async def future(self):
        self._file_upload_envelope = await FetchFileUploadS3Data(self._pubnub).\
            channel(self._channel).\
            file_name(self._file_name).future()

        response_envelope = await super(SendFileNative, self).future()

        publish_file_response = await PublishFileMessage(self._pubnub).\
            channel(self._channel).\
            meta(self._meta).\
            message(self._message).\
            file_id(response_envelope.result.file_id).\
            file_name(response_envelope.result.name).\
            should_store(self._should_store).\
            ttl(self._ttl).\
            cipher_key(self._cipher_key).future()

        response_envelope.result.timestamp = publish_file_response.result.timestamp
        return response_envelope

    async def result(self):
        response_envelope = await self.future()
        return response_envelope.result
